/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2007, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

cumulativesigma.c is for computing the cumulative of the normal
distribution. The sigma is the only command line argument, and the
value of the cumulative of the normal distribution for that sigma is
printed to stdout.

$Revision: 0.0 $
$Date: 2007/01/18 19:36:00 $
$Id: cumulativesigma.c,v 0.0 2007/01/18 19:36:00 john Exp $
$Log: cumulativesigma.c,v $
Revision 0.0  2007/01/18 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#ifdef __STDC__

#include <float.h>

#endif

#ifndef PI /* make sure PI is defined */

#define PI 3.14159265358979323846 /* pi to 20 decimal places */

#endif

#ifdef __STDC__

static char rcsid[] = "$Id: cumulativesigma.c,v 0.0 2007/01/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2007, John Conover, All Rights Reserved"; /* the copyright banner */

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Compute the cumulative of the normal distribution\n",
    "Usage: cumulativesigma -s sigma [-v]\n",
    "    -s sigma, the value to compute, (-5 to +4.99...)\n",
    "    -v, print the program's version information\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static void cumulativenormal (void); /* construct the cumulative of the normal distribution */

#else

static void print_message (); /* print any error messages */
static void cumulativenormal (); /* construct the cumulative of the normal distribution */

#endif

#define SIGMAS 5 /* 5 sigma limit, ie., -5 to +5 sigma */
#define STEPS_PER_SIGMA 10000 /* each sigma has 1000 steps of granularity */

static double cumulative[2 * SIGMAS * STEPS_PER_SIGMA]; /* the array of cumulative levels, ie., the cumulative of the normal distribution */

static int sigma_limit = 2 * SIGMAS * STEPS_PER_SIGMA; /* the array size of the array of cumulative levels, ie., 2 * SIGMAS * STEPS_PER_SIGMA, for calculation expediency */

static int cumulativeconstructed = 0; /* flag to determine whether the cumulative normal distribution array, cumulative[], has been set up, 0 = no, 1 = yes */

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    int c, /* command line switch */
        retval = EARGS; /* return value, assume not enough arguments */

    double sigma = (double) -(2 * SIGMAS); /* the sigma value to compute the cumulative of the normal distribution */

    while ((c = getopt (argc, argv, "s:v")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 's': /* request for sigma value? */

                sigma = atof (optarg); /* save the value of sigma */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    if ((sigma >= (double) -SIGMAS) && (sigma < (double) SIGMAS)) /* valid value of sigma? */
    {
        retval = NOERROR; /* assume no error */

        if (cumulativeconstructed == 0) /* cumulative normal distribution array, cumulative[], been set up? 0 = no, 1 = yes */
        {
            cumulativenormal (); /* no, set up the cumulative normal distribution array, cumulative[] */
        }

        (void) printf ("%f\n", cumulative[(int) floor (STEPS_PER_SIGMA * (SIGMAS + sigma))]); /* print the value of the cumulative of the normal distribution for this sigma */
    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

Construct the cumulative of the normal distribution.

static void cumulativenormal (void);

I) The process chosen uses a binary search of the cumulative of the
normal distribution. First a "graph" or "table" (ie., like the table
above,) is constructed in memory of the cumulative of the normal
distribution. This "graph" is then searched to compute the value of
the cumulative of the normal distribution.

II) Constructs the array, cumulative[], which is used as a lookup
mechanism for the cumulative of a normal distribution.

III) This is simply a lookup table, with the implicit index
representing a sigma value.  Each sigma is divided into
STEPS_PER_SIGMA many steps.

Returns nothing.

*/

#ifdef __STDC__

static void cumulativenormal (void)

#else

static void cumulativenormal ()

#endif

{
    int i; /* step counter through cumulative of normal distribution */

    double scale = (double) 1.0 / sqrt ((double) 2.0 * (double) PI), /* scaling factor for cumulative of normal distribution, for math expediency */
           del = (double) 1.0 / (double) STEPS_PER_SIGMA, /* increment of granularity of x axis of cumulative of normal distribution */
           x = (double) -SIGMAS, /* x variable in cumulative of normal distribution */
           cumulativesum = (double) 0.0; /* cumulative of normal distribution, begins at -5 sigma of the distribution, ie., x = -5 */

    for (i = 0; i < sigma_limit; i ++) /* for each step in the cumulative of the normal distribution */
    {
        cumulativesum = cumulativesum + ((scale * (exp ((- (x * x)) / ((double) 2.0)))) / (double) STEPS_PER_SIGMA); /* add the value of the normal distribution for this x to the cumulative of the normal distribution */
        cumulative[i] = cumulativesum; /* save the value of the cumulative of the normal distribution for this x */
        x = x + del; /* next increment of x in the cumulative of the normal distribution */
    }

    cumulativeconstructed = 1; /* set the flag to determine whether the cumulative normal distribution array, cumulative[], has been set up, 0 = no, 1 = yes */
}
